# -*- coding: utf-8 -*-
from indentation import Indentation
from statement import Statements

def default_indentation():
    return 4 * ' '

class SpecificationProcessor:
    def __init__(self, file_name):
        self.file_name = file_name

        self._statements = Statements()

        self._test_found = False
        self._long_comment_found = False

    def process(self, lines):
        ret = ['import unittest', 'import ' + self.file_name]

        # check file beginning now (attach defs as is)
        # this is a bit weak since it allows defs to be only at beginning
        defs = map(lambda a: a.startswith('def'), lines)

        def r_index(a, val):
            # missing list.rindex...
            return len(a) - a[::-1].index(val) - 1

        first_index = 0
        if any(defs):
            last_def_index = r_index(defs, True)

            first_index = map(lambda a: a.strip(), lines[last_def_index:]).index('')
            first_index += last_def_index

        processed_lines = map(lambda a: self.process_line(a), lines[:first_index])
        ret.extend(processed_lines)

        test_class_name = 'Test' + self.file_name.capitalize()
        ret.append('class ' + test_class_name + '(unittest.TestCase):')

        new_lines, set_up = self.pick_set_up(lines[first_index:])

        for line in new_lines:
            processed_line = self.process_line(line, set_up)
            
            if processed_line:
                ret.append(processed_line)
        
        ret.extend(['suite = unittest.TestLoader().loadTestsFromTestCase(' + \
            test_class_name + ')',
            'unittest.TextTestRunner(verbosity=2).run(suite)'
        ])

        return '\n'.join(ret)

    def pick_set_up(self, lines):
        found_set_up = False
        new_lines = []
        set_up = []

        for line in lines:
            if line[0] == ' ':
                if found_set_up:
                    set_up.append(line)
                else:
                    new_lines.append(line)
            else:
                stripped_line = line.strip()
                if len(stripped_line) > 0:
                    if stripped_line == 'set up':
                        found_set_up = True
                    else:
                        new_lines.append(line)
                elif found_set_up:
                    found_set_up = False
                else:
                    new_lines.append(line)

        return new_lines, set_up

    def process_line(self, line, set_up=()):
        stripped_line = line.strip()

        if len(stripped_line) > 0 and stripped_line[0] == '#':
            return line

        skips = ('def', 'return')
        if any(map(lambda a: stripped_line.startswith(a + ' '), skips)):
            return line

        just_found = False
        if stripped_line.endswith("'''"):
            self._long_comment_found = '=' in stripped_line
            just_found = True

            if not self._long_comment_found:
                return line

        if not just_found and self._long_comment_found:
            return line

        if line and line[0] == ' ':
            indentation = Indentation(line)
            ret = self._statements.convert(stripped_line)

            if hasattr(ret, '__iter__'):
                return default_indentation() + indentation() + ret[0] + \
                    '\n' + default_indentation() + indentation() + ret[1]

            return (default_indentation() + indentation() + ret).rstrip()

        if len(stripped_line) > 0 and not self._test_found:
            self._test_found = True

            ret = '\n' + default_indentation() + 'def test_' + \
                stripped_line.replace(' ', '_') + '(self):'

            for content in set_up:
                ret += '\n' + default_indentation() + content

            return ret

        if not stripped_line or stripped_line == '\n':
            self._test_found = False

        return (default_indentation() + line).rstrip()



# -*- coding: utf-8 -*-
class Statement:
    def matches(self, line):
        return self.value in line

    def convert(self, line):
        parts = line.split(self.value)
        parts = [part.strip() for part in parts]
        parts_len = len(parts)
        parts_lim = int(round(parts_len / 2.0))

        l_part = self.value.join(parts[0:parts_lim])
        r_part = self.value.join(parts[parts_lim:parts_len])

        code_params = self._code_parameters(l_part, r_part)
        return 'self.' + self.code + '(' + code_params + ')'

    def _code_parameters(self, l_part, r_part):
        return l_part + ', ' + r_part

class Equals(Statement):
    value = '=='
    code = 'assertEqual'

class NotEquals(Statement):
    value = '!='
    code = 'assertNotEqual'

class AlmostEquals(Statement):
    value = '~='
    code = 'assertAlmostEqual'

class AlmostNotEquals(Statement):
    value = '!~='
    code = 'assertNotAlmostEqual'

class Raises(Statement):
    value = 'raises'

    def convert(self, line):
        expr, error = line.split('raises')

        return ['try:' + expr, 'except' + error + ': pass']

class Inequality(Statement):
    code = 'assertTrue'

    def _code_parameters(self, l_part, r_part):
        l_parts = l_part.split(self.value)

        return (' ' + self.value + ' ').join(l_parts) + ' ' + self.value + \
            ' ' + r_part

class BiggerThan(Inequality):
    value = '>'

class BiggerThanOrEquals(Inequality):
    value = '>='

class SmallerThan(Inequality):
    value = '<'

class SmallerThanEquals(Inequality):
    value = '<='

class Any(Statement):
    def matches(self, line):
        return True

    def convert(self, line):
        return line

class Statements(list):
    def __init__(self):
        statements = (Equals(), NotEquals(), AlmostNotEquals(),
            AlmostEquals(), BiggerThanOrEquals(), BiggerThan(),
            SmallerThanEquals(), SmallerThan(), Raises(), Any())

        super(Statements, self).__init__(statements)

    def convert(self, line):
        for statement in self:
            if statement.matches(line):
                return statement.convert(line)



# -*- coding: utf-8 -*-
from __future__ import with_statement

import glob
import imp
import os
import sys
import tempfile
import time
from optparse import OptionParser
from processor import SpecificationProcessor
from itertools import chain

import __init__


def get_base_name(a):
    return os.path.splitext(a)[0]

def run(spec_files):
    for spec_file_name in spec_files:
        base_name = get_base_name(spec_file_name)
        py_file_name = base_name + '.py'

        if os.path.exists(py_file_name):
            print('\n**Testing ' + spec_file_name + '**\n')

            processor = SpecificationProcessor(base_name)

            with open(spec_file_name) as f:
                lines = f.readlines()

            spec_code = processor.process(lines)

            # http://docs.python.org/library/tempfile.html#tempfile.mktemp
            tmp_file = tempfile.NamedTemporaryFile(delete=False, suffix='.py')
            tmp_file.write(spec_code)
            tmp_file.close()

            try:
                try:
                    module = imp.load_source('spec', tmp_file.name)
                except SystemExit:
                    print('Module ' + base_name + ' not found!');
            except Exception, e:
                os.unlink(tmp_file.name)
                sys.exit(str(e))

            os.unlink(tmp_file.name)
            os.unlink(tmp_file.name + 'c')

def get_specs():
    return glob.glob('*.spec')

def output_tests(option, opt, output_dir, parser):
    got_all = False

    for spec_name in get_specs():
        with open(spec_name) as f:
            lines = f.readlines()

        base_name = get_base_name(spec_name)
        processor = SpecificationProcessor(base_name)
        spec_code = processor.process(lines)

        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

        py_file_name = 'test_' + base_name + '.py'
        py_file = os.path.join(output_dir, py_file_name)

        def write_file():
            print('Writing ' + py_file)

            with open(py_file, 'w') as f:
                f.write(spec_code)

        if os.path.exists(py_file):
            def do_nothing():
                print('Doing nothing.')

            possible_answers = {
                'Y': write_file,
                'N': do_nothing,
                'A': None
            }
            opts = '/'.join(possible_answers.keys())
            answer = None

            while answer not in possible_answers:
                if got_all:
                    answer = 'Y'
                else:
                    answer = raw_input('Are you sure you want to override file (' +
                        py_file_name + ')?\n' + opts + '\n')
                    answer = answer.upper()

                    if answer == 'A':
                        got_all = True
                        answer = 'Y'

            possible_answers[answer]()
        else:
            write_file()

def run_tests(spec_files):
    if len(spec_files) == 0:
        print('\nNo specifications found!')

        return False
    else:
        print('\nRunning tests')

        run(spec_files)

        return True

def looping_run(*args):
    def get_file_times(filenames):
        get_file_time = lambda filename: os.stat(filename).st_mtime

        return map(get_file_time, filenames)

    specs = get_specs()
    tests_found = run_tests(specs)

    if not tests_found:
        return

    file_times = get_file_times(specs)

    while True:
        new_file_times = get_file_times(specs)

        if file_times != new_file_times:
            file_times = new_file_times

            run_tests(specs)

        time.sleep(1)

def show_version(*args):
    print("speccer %s" % __init__.__version__)
    sys.exit(0)

def main():
    # make sure current working directory is in the path
    # this is needed for imports to work
    cwd = os.getcwd()
    sys.path.append(cwd)

    usage = """usage: %prog command [args] [options]"""

    description = """Brief Description:
Speccer is a specification based test runner."""

    epilog = """\nLong Description:
Speccer is a specification based test runner. It uses a special syntax to wrap
Python's native unittest module. Tests are defined like this:

myclass.spec (tests myclass.py):
set up
    c = myclass.MyClass()

adds two and two
    c.add(2,2) == 4

adds negatives
    c.add(10, -10) == 0

fails adding int and string
    c.add(10, 'foo') raises TypeError

As you can see the syntax is quite light. It's pretty much like regular Python
but with less noise. Asserts and function definitions are implicit.

In order to run the tests, just run the tool without any parameters in the
folder they are in."""

    class MyParser(OptionParser):
        def format_epilog(self, formatter):
            return self.epilog

    parser = MyParser(usage=usage, description=description, epilog=epilog)
    parser.add_option("-v", "--version", action="callback",
        callback=show_version,
        help="show program's version number and exit")
    parser.add_option("-l", "--loop", action="callback",
        callback=looping_run,
        help="run tests in a looping manner. Tests get run each time a spec file is changed")
    parser.add_option("-o", "--output", action="callback",
        dest="output_folder", type="string",
        callback=output_tests,
        help="output generated test files to given folder")

    options, args = parser.parse_args()

    if len(sys.argv) < 2:
        run_tests(get_specs())
    elif len(args) > 0:
        # TODO: make Python imports work with relative paths in order to enable this
        # expand dir contents to args
        #args = list(chain(*map(lambda a: map(lambda b: os.path.join(a, b), os.listdir(a)) if os.path.isdir(a) else a, args)))

        # get rid of files end with other than .spec or not containing extension at all
        args = filter(lambda a: a.endswith('.spec') or a.find('.') == -1, args)

        # add .spec to files that don't have it yet
        args = map(lambda a: a + '.spec' if not a.endswith('.spec') else a, args)

        # make sure files really exist
        args = filter(lambda a: os.path.exists(a), args)

        run_tests(args)

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
class Indentation:
    def __init__(self, line):
        indentation_end = self._find_last_from_beginning(line, ' ')
        self.amount = indentation_end + 1

    def _find_last_from_beginning(self, line, needle):
        for i, char in enumerate(line):
            if char != needle:
                return i - 1

        return 0

    def __call__(self):
        return self.amount * ' '



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import subprocess
import speccer
from setuptools import setup

def pandoc(source, from_format, to_format):
    # http://osiux.com/html-to-restructured-text-in-python-using-pandoc
    # raises OSError if pandoc is not found!
    p = subprocess.Popen(['pandoc', '--from=' + from_format, '--to=' + to_format],
        stdin=subprocess.PIPE, stdout=subprocess.PIPE
        )
    return p.communicate(source)[0]

description = "Specification based test runner."
try:
    md = open('README.md').read()

    long_description = pandoc(md, 'markdown', 'rst')
except (IOError, OSError):
    print 'check that you have installed pandoc properly and that README.md exists!'
    long_description = description

setup(
    name = "speccer",
    version = speccer.__version__,
    url = 'https://github.com/bebraw/speccer',
    license = 'MIT',
    description = description,
    long_description = long_description,
    author = speccer.__author__,
    author_email = 'bebraw@gmail.com',
    packages = ['speccer', ],
    package_dir = {'speccer': 'speccer', },
    install_requires = ['setuptools', ],
    entry_points="""
    [console_scripts]
    speccer = speccer.runner:main
    """,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Topic :: Software Development :: Testing',
    ],
)
